pub mod is_prime;

use std::{time::SystemTime,thread};
use num_bigint::BigUint;
use crate::is_prime::prime_utils;
use thread_priority::{set_current_thread_priority,ThreadPriority};

fn p(n: f64) -> f64 {
    let ln_n = n.ln();
    let ln_ln_n = ln_n.ln();
    let ln_ln_ln_n = ln_ln_n.ln();

    n * (
        ln_n + ln_ln_n - 1.0
        + (ln_ln_ln_n - 2.0) / ln_n
        - ((ln_ln_n).powi(2) - 6.0 * ln_ln_n + 11.0) / (2.0 * 2.0f64.log2() * ln_n)
        + ((ln_ln_n / ln_n).powi(3)) * (1.0 / ln_n)
    )
}


fn main() {

    assert!(set_current_thread_priority(ThreadPriority::Max).is_ok());

    /* 
    let mut buffer = String::new();
    let stdin = io::stdin();
    print!("please enter the number to check: ");
    io::stdout().flush().expect("could not flush");
    stdin.read_line(&mut buffer).expect("could not read from stdin");
    buffer = buffer.trim().to_owned();
    
    let number = BigUint::parse_bytes(&buffer.as_bytes(), 10).expect("expected positive number");

    if is_prime(number.clone()) {
        println!("\n{number} is a prime");
    } else {
        println!("\n{number} is not a prime");
    }
    */

    let sys_time = SystemTime::now();

    let n:f64 = 1_000_000f64;
    let nthprime_approx = p(n*1.02) as u128;
    const THREADS: u128 = 12;
    let workloads = nthprime_approx/THREADS;

    let primecache: Vec<BigUint> = vec![];
    let primemutex = std::sync::Mutex::new(primecache);
    let primearc = std::sync::Arc::new(primemutex);
    let mut joins = vec![];
    for i in 0..THREADS {
        let arcclone = primearc.clone();
        let tprimes = thread::spawn(move || {

            assert!(set_current_thread_priority(ThreadPriority::Max).is_ok());

            let one_thousand = &BigUint::from(1_000u16);
            let one_hundred = &BigUint::from(100u8);
            let ten = &BigUint::from(10u8);
            

            let work_uint = BigUint::from(workloads);
            let mut number = BigUint::from(i)*&work_uint;
            let inc = BigUint::from(1u8);
            // let zero = BigUint::from(0u8);
            
            let max = BigUint::from(i+1)*work_uint;
            let mut primecache = vec![];
            loop {
                number = number + &inc;
                
                if prime_utils::is_prime(&number,&mut primecache) {
                    primecache.push(number.clone());
                    //println!("Found a prime {} {}",number, primecache.len());
                }
                if (&number * one_thousand)/&max/ten == &number/&max*one_hundred {
                    let mut new_primecache = arcclone.lock().unwrap();
                    new_primecache.append(&mut primecache);
                    new_primecache.sort();
                    new_primecache.dedup();
                    
                    primecache = new_primecache.clone();
                    
                }
                if number == max {
                    break;
                }
            }   

            let mut new_primecache = arcclone.lock().unwrap();
            new_primecache.append(&mut primecache);
            drop(primecache);
            new_primecache.sort();
            new_primecache.dedup();

            true
        });
        joins.push(tprimes);
    }

    loop {
        let mut njoins = vec![];
        for join in joins {
            if join.is_finished() {
                join.join().unwrap();
            } else {
                njoins.push(join);
            }
        }
        joins = njoins;
        if joins.len()==0 {
            break;
        }
    }

    let mut primes = primearc.lock().unwrap().to_vec();

    let removed = (primes.len()-n as usize).clamp(0, primes.len());
    if removed != 0 {
        println!("removing {removed} primes");
        for _ in 0..removed {
            primes.pop();
        }
    }

    let added = n as usize-primes.len();
    if primes.len() < n as usize {
        println!("Less primes than expected!");
        let mut number = BigUint::from(THREADS)*BigUint::from(workloads);
        let inc = BigUint::from(1u8);
        loop {
            number = number + &inc;
            if prime_utils::is_prime(&number, &primes) {
                primes.push(number.clone());
            }
            if primes.len() == n as usize {
                break;
            }
        }
    }

    let new_sys_time = SystemTime::now();
    let difference = new_sys_time.duration_since(sys_time)
    .expect("Clock may have gone backwards");

    println!("{:?}", primes);

    println!("prime count: {}, expected: {}, primes added: {} | {}%, removed: {}",primes.len(),n,added,(added as f64)/primes.len() as f64*100.0, removed);
    println!("last prime approx: {}",nthprime_approx);

    println!("And that's time: {difference:?}");


}
